import { isInput, isNonNullType, isListType, isNamedType, ObjectTypeDefinitionBuilder } from './../graphql';
import { DeclarationBlock, indent } from '@graphql-codegen/visitor-plugin-common';
import { Visitor } from '../visitor';
import { buildApi, formatDirectiveConfig } from '../directive';
const importZod = `import * as myzod from 'myzod'`;
const anySchema = `definedNonNullAnySchema`;
export const MyZodSchemaVisitor = (schema, config) => {
    const importTypes = [];
    return {
        buildImports: () => {
            if (config.importFrom && importTypes.length > 0) {
                return [importZod, `import { ${importTypes.join(', ')} } from '${config.importFrom}'`];
            }
            return [importZod];
        },
        initialEmit: () => '\n' +
            [
                new DeclarationBlock({}).export().asKind('const').withName(`${anySchema}`).withContent(`myzod.object({})`)
                    .string,
            ].join('\n'),
        InputObjectTypeDefinition: {
            leave: (node) => {
                const visitor = new Visitor('input', schema, config);
                const name = visitor.convertName(node.name.value);
                importTypes.push(name);
                const shape = node.fields?.map(field => generateFieldMyZodSchema(config, visitor, field, 2)).join(',\n');
                return new DeclarationBlock({})
                    .export()
                    .asKind('function')
                    .withName(`${name}Schema(): myzod.Type<${name}>`)
                    .withBlock([indent(`return myzod.object({`), shape, indent('})')].join('\n')).string;
            },
        },
        ObjectTypeDefinition: {
            leave: ObjectTypeDefinitionBuilder(config.withObjectType, (node) => {
                const visitor = new Visitor('output', schema, config);
                const name = visitor.convertName(node.name.value);
                importTypes.push(name);
                const shape = node.fields?.map(field => generateFieldMyZodSchema(config, visitor, field, 2)).join(',\n');
                return new DeclarationBlock({})
                    .export()
                    .asKind('function')
                    .withName(`${name}Schema(): myzod.Type<${name}>`)
                    .withBlock([
                    indent(`return myzod.object({`),
                    indent(`__typename: myzod.literal('${node.name.value}').optional(),`, 2),
                    shape,
                    indent('})'),
                ].join('\n')).string;
            }),
        },
        EnumTypeDefinition: {
            leave: (node) => {
                const visitor = new Visitor('both', schema, config);
                const enumname = visitor.convertName(node.name.value);
                importTypes.push(enumname);
                // z.enum are basically myzod.literals
                if (config.enumsAsTypes) {
                    return new DeclarationBlock({})
                        .export()
                        .asKind('type')
                        .withName(`${enumname}Schema`)
                        .withContent(`myzod.literals(${node.values?.map(enumOption => `'${enumOption.name.value}'`).join(', ')})`)
                        .string;
                }
                return new DeclarationBlock({})
                    .export()
                    .asKind('const')
                    .withName(`${enumname}Schema`)
                    .withContent(`myzod.enum(${enumname})`).string;
            },
        },
        UnionTypeDefinition: {
            leave: (node) => {
                if (!node.types || !config.withObjectType)
                    return;
                const visitor = new Visitor('output', schema, config);
                const unionName = visitor.convertName(node.name.value);
                const unionElements = node.types
                    ?.map(t => {
                    const element = visitor.convertName(t.name.value);
                    const typ = visitor.getType(t.name.value);
                    if (typ?.astNode?.kind === 'EnumTypeDefinition') {
                        return `${element}Schema`;
                    }
                    return `${element}Schema()`;
                })
                    .join(', ');
                const unionElementsCount = node.types?.length ?? 0;
                const union = unionElementsCount > 1 ? indent(`return myzod.union([${unionElements}])`) : indent(`return ${unionElements}`);
                return new DeclarationBlock({}).export().asKind('function').withName(`${unionName}Schema()`).withBlock(union)
                    .string;
            },
        },
    };
};
const generateFieldMyZodSchema = (config, visitor, field, indentCount) => {
    const gen = generateFieldTypeMyZodSchema(config, visitor, field, field.type);
    return indent(`${field.name.value}: ${maybeLazy(field.type, gen)}`, indentCount);
};
const generateFieldTypeMyZodSchema = (config, visitor, field, type, parentType) => {
    if (isListType(type)) {
        const gen = generateFieldTypeMyZodSchema(config, visitor, field, type.type, type);
        if (!isNonNullType(parentType)) {
            const arrayGen = `myzod.array(${maybeLazy(type.type, gen)})`;
            const maybeLazyGen = applyDirectives(config, field, arrayGen);
            return `${maybeLazyGen}.optional().nullable()`;
        }
        return `myzod.array(${maybeLazy(type.type, gen)})`;
    }
    if (isNonNullType(type)) {
        const gen = generateFieldTypeMyZodSchema(config, visitor, field, type.type, type);
        return maybeLazy(type.type, gen);
    }
    if (isNamedType(type)) {
        const gen = generateNameNodeMyZodSchema(config, visitor, type.name);
        if (isListType(parentType)) {
            return `${gen}.nullable()`;
        }
        const appliedDirectivesGen = applyDirectives(config, field, gen);
        if (isNonNullType(parentType)) {
            if (config.notAllowEmptyString === true) {
                const tsType = visitor.getScalarType(type.name.value);
                if (tsType === 'string')
                    return `${gen}.min(1)`;
            }
            return appliedDirectivesGen;
        }
        if (isListType(parentType)) {
            return `${appliedDirectivesGen}.nullable()`;
        }
        return `${appliedDirectivesGen}.optional().nullable()`;
    }
    console.warn('unhandled type:', type);
    return '';
};
const applyDirectives = (config, field, gen) => {
    if (config.directives && field.directives) {
        const formatted = formatDirectiveConfig(config.directives);
        return gen + buildApi(formatted, field.directives);
    }
    return gen;
};
const generateNameNodeMyZodSchema = (config, visitor, node) => {
    const converter = visitor.getNameNodeConverter(node);
    if (converter?.targetKind === 'InputObjectTypeDefinition') {
        const name = converter.convertName();
        return `${name}Schema()`;
    }
    if (converter?.targetKind === 'ObjectTypeDefinition') {
        const name = converter.convertName();
        return `${name}Schema()`;
    }
    if (converter?.targetKind === 'EnumTypeDefinition') {
        const name = converter.convertName();
        return `${name}Schema`;
    }
    if (converter?.targetKind === 'UnionTypeDefinition') {
        const name = converter.convertName();
        return `${name}Schema()`;
    }
    return myzod4Scalar(config, visitor, node.value);
};
const maybeLazy = (type, schema) => {
    if (isNamedType(type) && isInput(type.name.value)) {
        return `myzod.lazy(() => ${schema})`;
    }
    return schema;
};
const myzod4Scalar = (config, visitor, scalarName) => {
    if (config.scalarSchemas?.[scalarName]) {
        return config.scalarSchemas[scalarName];
    }
    const tsType = visitor.getScalarType(scalarName);
    switch (tsType) {
        case 'string':
            return `myzod.string()`;
        case 'number':
            return `myzod.number()`;
        case 'boolean':
            return `myzod.boolean()`;
    }
    console.warn('unhandled name:', scalarName);
    return anySchema;
};
